#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include "marketfunc.h"

#define NB_ACTION_TYPE 1
#define MAX_TIME 5
#define NB_CLIENT_MAX 5
#define NB_MEMORISATION_MAX 20
#define ARGENT_INIT 1000


struct client{
  	int id;
 	int argent;
 	int actionPossede[NB_ACTION_TYPE];
};

struct compteRendu ordreSell(struct order ordre);
struct compteRendu ordreBuy(struct order ordre);
int envoyerCompteRendu(struct compteRendu cr);
void afficherEtat(int nbClient, struct client listeClient[]);



struct compteRendu{
	int type;
	int nbAction;
	int prix;
	int pidDest;
};


void handlerAlrm(int a)
{
	printf("SERVER : Rien n'a été recu depuis un certain temps... STOP\n");
	unlink("pipe_marketServer");
	marketfunc_terminate();
	exit(0);
}

int main(int argc, char * argv[])
{
	int error;
	int tube;
	struct order ordre;
	struct client listeClient[NB_CLIENT_MAX];
	struct order memoireSeuil[NB_MEMORISATION_MAX];
	struct compteRendu cr;
	int nbClient = 0;
	int nbSeuil = 0;
	int val;
	
	printf("SERVER : Lancement du serveur\n");
	
	// Mise en place du handler pour le signal SIGALRM
	signal(SIGALRM, handlerAlrm);
	
	printf("signal ok \n");
	// ETAPE 1 : INITIALISATION DE LA BIBLIOTHEQUE
	error = marketfunc_init(NB_ACTION_TYPE);
	if (error != 0)
	{
		perror("Echec lors de l'initialisation de la bibliothèque ...");
		exit(-1);
	}
	
	unlink("pipe_marketServer");
	// ETAPE 2 : CREATION ET OUVERTURE DU TUBE NOMME
	if (mkfifo("pipe_marketServer", 0644) < 0)
	{
		perror("Echec lors de la creation du tube nommé ...");
		exit(-2);
	}
	
	printf("Ouverture du tube \n");
	tube = open("pipe_marketServer", O_RDONLY);
	
	alarm(MAX_TIME);
	
	while(1)
	{
		sleep(1);
		// ETAPE 3 : LECTURE DANS LE TUBE DES ORDRES PUIS TRAITEMENT
		printf("EN attente lecture \n");

		read(tube, &ordre, sizeof(ordre));
		
		// On a réussi a lire, mise a jour de alarm
		alarm (MAX_TIME);
		
		switch(ordre.type)
		{
			case OT_REGISTER :
				printf("SERVER : Enregistrement d'un nouveau client en cours ...\n");
				
				// Ajouter le client a la liste
				struct client unClient;
				unClient.id = ordre.sender;
				unClient.argent = ARGENT_INIT;
				unClient.actionPossede[0] = 0;
				
				if(nbClient <= NB_CLIENT_MAX)
				{
					listeClient[nbClient] = unClient;
					nbClient++;
			
					cr.nbAction = get_action_quantity(0);
					cr.type = 0;
					cr.prix = get_price(0);
					cr.pidDest = ordre.sender;
			
					printf("Envoi du compte rendu\n");
					envoyerCompteRendu(cr);
					printf("SERVER : Enregistrement d'un nouveau client terminé ...\n");
					
				}
				else
				{
					// Erreur, la liste est pleine
					printf("SERVER : Impossible d'ajouter un nouveau client.");
				}
			
			break;
		
			case OT_ALARM_HIGH :
				printf("SERVER : Mémorisation d'un seuil maximum en cours ...\n");
				// Enregistrement du seuil d'alarme
				if(nbSeuil <= NB_MEMORISATION_MAX)
				{
					memoireSeuil[nbSeuil] = ordre;
					nbSeuil++;
				}
				else
				{
					printf("SERVER : La memoire des seuils est pleine ...\n");
				}
			break;
			
			case OT_ALARM_LOW :
				printf("SERVER : Mémorisation d'un seuil minimum en cours en cours ...\n");
				// Enregistrement du seuil d'alarme
				if(nbSeuil <= NB_MEMORISATION_MAX)
				{
					memoireSeuil[nbSeuil] = ordre;
					nbSeuil++;
				}
				else
				{
					printf("SERVER : La memoire des seuils est pleine ...\n");
				}
			break;
			
			case OT_BUY :
				cr = ordreBuy(ordre);
				// Mise a jour de l'etat du serveur
				int i;
				for(i=0; i<= nbClient; i++)
				{
					if(listeClient[i].id == ordre.sender)
					{
						listeClient[i].argent -= cr.prix;
						listeClient[i].actionPossede[0] += cr.nbAction;
					} 
				}
				envoyerCompteRendu(cr);
			break;
			case OT_SELL :
				cr = ordreSell(ordre);
				for(i=0; i<= nbClient; i++)
				{
					if(listeClient[i].id == ordre.sender)
					{
						listeClient[i].argent += cr.prix;
						listeClient[i].actionPossede[0] -= cr.nbAction;
					} 
				}
				envoyerCompteRendu(cr);
			break;
		
			case OT_REQUEST_PRICE :
				printf("SERVER : Requete sur le prix d'une action en cours ...\n");
				val = get_price(ordre.val1);
				// Envi du prix
			
				cr.type = ordre.val1;
				cr.prix = val;
				cr.pidDest = ordre.sender;
				
				envoyerCompteRendu(cr);
				
			break;
		}
		
		afficherEtat(nbClient, listeClient);
	

		// ETAPE 4 : MISE A JOUR DES PRIX ET ENVOI DES SIGNAUX AUX ACTORS
		int i;
		for(i=0; i<= nbSeuil; i++)
		{
			// OT_ALARM_HIGH
			if(memoireSeuil[i].type == 2)
			{
				if(get_price(memoireSeuil[i].type) > memoireSeuil[i].val1)
				{
					printf("ENVOI D'UN SIGNAL !\n");
					kill(SIGUSR1, memoireSeuil[i].sender);
				}
			}
			
			 // OT_ALARM_LOW
			if(memoireSeuil[i].type == 3)
			{
				if(get_price(memoireSeuil[i].type) < memoireSeuil[i].val1)
				{
					printf("ENVOI D'UN SIGNAL !\n");
					kill(SIGUSR2, memoireSeuil[i].sender);
				}
			}
		}
	}
}

// Gere l'action d'achat puis retourne un compte rendu
struct compteRendu ordreBuy(struct order ordre)
{
	int nbAction = ordre.val1;
	int prix = ordre.val2;
	int i;
	int nb = 0;
	int coutTotal = 0;
	struct compteRendu cr;
	
	for(i=0; i<nbAction; i++)
	{
		int ret;
		printf("Buy\n");
		ret = buy(0, prix);
		
		if(ret != 0)
		{
			nb++;
			coutTotal += ret;
		}
	}

	cr.type = ordre.type;
	cr.nbAction = nb;
	cr.prix = coutTotal;
	cr.pidDest = ordre.sender;
	
	return cr;
}

// Gere l'action de vente, puis retourne un compte rendu
struct compteRendu ordreSell(struct order ordre)
{
	int nbAction = ordre.val1;
	int prix = ordre.val2;
	int i;
	int nb = 0;
	int coutTotal = 0;
	struct compteRendu cr;
	
	for(i=0; i<nbAction; i++)
	{
		int ret;
		printf("Sell\n");
		ret = sell(0, prix);
		
		if(ret != 0)
		{
			nb++;
			coutTotal += ret;
		}
	}

	cr.type = ordre.type;
	cr.nbAction = nb;
	cr.prix = coutTotal;
	cr.pidDest = ordre.sender;
	
	return cr;
}

// Envoi un compte rendu a l'actor qui a envoyer l'order qui y correspond
int envoyerCompteRendu(struct compteRendu cr)
{
	char * nomPipe = malloc(sizeof(char) * 50);
	sprintf(nomPipe,"pipe_client_%d",cr.pidDest);
	int tubeActor = open(nomPipe, O_WRONLY);
	printf("Ppen OK \n");

	printf("Voici le CR a envoyer : Type = %d, nbAction = %d, prix = %d\n", cr.type, cr.nbAction, cr.prix);
	write(tubeActor, &cr, sizeof(cr));
	
	close(tubeActor);
	return 0;
}

// Affiche l'etat du serveur (Nb client + pour chaque client, le nombre d'action qu'il possede)
void afficherEtat(int nbClient, struct client listeClient[])
{
	printf("*** ETAT DU SERVEUR\n");
	printf("*** Je connais actuellement %d clients\n", nbClient);
	int i;
	for(i=0; i<=nbClient-1; i++)
	{
		printf("Client %d : ID = %d\n", i, listeClient[i].id);
		int j = 0;
		for(j=0; j<= NB_ACTION_TYPE -1; j++)
		{
			printf("Client %d possède %d actions de type %d\n", i, listeClient[i].actionPossede[j], j);
		}
	}
}

